#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include "tron.h"
#include <time.h>

#ifndef min
template <class T> static inline T min(T x,T y) { return (x<y)?x:y; }
#endif

#ifndef max
template <class T> static inline T max(T x,T y) { return (x>y)?x:y; }
#endif

#ifdef __cplusplus
extern "C" {
#endif

extern double dnrm2_(int *, double *, int *);
extern double ddot_(int *, double *, int *, double *, int *);
extern int daxpy_(int *, double *, double *, int *, double *, int *);
extern int dscal_(int *, double *, double *, int *);

#ifdef __cplusplus
}
#endif

static void default_print(const char *buf)
{
	fputs(buf,stdout);
	fflush(stdout);
}


double dnrm1_(int *n, double *x, int *incx)
{
  long int ix, nn, iincx;
  double norm, scale, absxi, ssq, temp;

  // Dereference inputs 
  nn = *n;
  iincx = *incx;

  if( nn > 0 && iincx > 0 )
  {
    if (nn == 1)
    {
      norm = fabs(x[0]);
    }
    else
    {
      scale = 0.0;
      ssq = 1.0;
      for (ix=(nn-1)*iincx; ix>=0; ix-=iincx)
      {
        if (x[ix] != 0.0)
        {
          absxi = fabs(x[ix]);
          if (scale < absxi)
          {
            temp = scale / absxi;
            ssq = ssq * temp + 1.0;
            scale = absxi;
          }
          else
          {
            temp = absxi / scale;
            ssq += temp;
          }
        }
      }
      norm = scale * ssq;
    }
  }
  else
    norm = 0.0;
  return norm;
}

void TRON::info(const char *fmt,...)
{
	char buf[BUFSIZ];
	va_list ap;
	va_start(ap,fmt);
	vsprintf(buf,fmt,ap);
	va_end(ap);
	(*tron_print_string)(buf);
}

TRON::TRON(const function *fun_obj, double eps, double eps_cg, int max_iter)
{
	this->fun_obj=const_cast<function *>(fun_obj);
	this->eps=eps;
	this->eps_cg=eps_cg;
	this->max_iter=max_iter;
	tron_print_string = default_print;
}

TRON::~TRON()
{
}


void TRON::tron(double *w, const parameter *param)
{
	int n = fun_obj->get_nr_variable();
	int i, cg_iter;
	double step_size;
	double f, fnew;
	double init_step_size = 1;
	int search = 1, iter = 1, inc = 1;
	double *s = new double[n];
	double *r = new double[n];
	double *g = new double[n];

        const double alpha_pcg = 0.01;
        double *M = new double[n];

	// calculate gradient norm at w=0 for stopping condition.
	double *w0 = new double[n];
	for (i=0; i<n; i++)
		w0[i] = 0;
	fun_obj->fun(w0);
	fun_obj->grad(w0, g);
	double gnorm0 = dnrm2_(&n, g, &inc);
	delete [] w0;
	f = fun_obj->fun(w);
	fun_obj->grad(w, g);
	double gnorm = dnrm2_(&n, g, &inc);

	if (gnorm <= eps*gnorm0)
		search = 0;

	iter = 1;

        if (param->prec == 1){
            fun_obj->get_diag_preconditioner(M);
            for(i=0; i<n; i++)
                    M[i] = (1-alpha_pcg) + alpha_pcg*M[i];
        }
        clock_t old_time = clock();
	while (iter <= max_iter && search)
	{
                if (param->prec == 1)
		    cg_iter = pcg(g, M, s, r, param);
                else
                    cg_iter = cg(g, s, r, param);

		step_size = fun_obj->line_search(s, w, g, init_step_size, &fnew);

		if (step_size == 0)
		{
			info("WARNING: line search fails\n");
			break;
		}

		daxpy_(&n, &step_size, s, &inc, w, &inc);
                clock_t new_time = clock();
                info("iter %2d f %5.10e |g| %5.3e CG %3d step_size %4.2e time %f \n", iter, f, gnorm, cg_iter, step_size, (float(new_time - old_time)/CLOCKS_PER_SEC));
                old_time = clock();

		f = fnew;
		iter++;

		fun_obj->grad(w, g);
                if (param->prec == 1){
                    fun_obj->get_diag_preconditioner(M);
		    for(i=0; i<n; i++)
		        M[i] = (1-alpha_pcg) + alpha_pcg*M[i];
                }

		gnorm = dnrm2_(&n, g, &inc);
		if (gnorm <= eps*gnorm0)
			break;
		if (f < -1.0e+32)
		{
			info("WARNING: f < -1.0e+32\n");
			break;
		}
	}

	delete[] g;
	delete[] r;
	delete[] s;
        delete[] M;
}

int TRON::stopping(int n, const parameter *param, double *g, double *r, int inc, int cg_iter, double new_Q, double Q, double *z, double Minv_dot_g)
{
    double ratio, forcing_seq;
    switch(param->ratio_type)
    {
        case RES:
            //residual
            if(param->prec == 0)
                ratio = dnrm2_(&n, r, &inc) / dnrm2_(&n, g, &inc);
            else
                ratio = sqrt(ddot_(&n, z, &inc, r, &inc)) / Minv_dot_g;
            break;
        case RESL1:
            //residual_l1
            ratio = dnrm1_(&n, r, &inc) / dnrm1_(&n, g, &inc);
            break;
        case QUAD:
            //quadratic
            ratio = cg_iter * (new_Q - Q) / (new_Q);
            break;
        default:
            info("Ratio type error!! Given: %d. Using instead Quadratic", param->ratio_type);
            ratio = cg_iter * (new_Q - Q) / (new_Q);
            break;
    }
    switch(param->forcing_seq_type)
    {
        case CONST:
            //constant
            forcing_seq = param->c0;
            break;
        case ADA:
            //adaptive
            forcing_seq = min(param->c1, param->c2*pow(dnrm2_(&n, g, &inc),param->c3));
            break;
        case ADAL1:
            //adaptive_l1
            forcing_seq = min(param->c1, param->c2*pow(dnrm1_(&n, g, &inc),param->c3));
            break;
        default:
            info("Forcing_seq type error!! Given: %d. Using instead Adaptive", param->forcing_seq_type);
            forcing_seq = min(param->c1, param->c2*pow(dnrm2_(&n, g, &inc),param->c3));
            break;
    }

    if (ratio <= forcing_seq)
       return 1;
    else
       return 0;
}

int TRON::cg(double *g, double *s, double *r, const parameter *param)
{
	int i, inc = 1;
	int n = fun_obj->get_nr_variable();
	double one = 1;
	double *d = new double[n];
	double *Hd = new double[n];
	double rTr, rnewTrnew, alpha, beta;

        double new_Q, Q = 0., epsilon = 0.0001;

	for (i=0; i<n; i++)
	{
		s[i] = 0;
		r[i] = -g[i];
		d[i] = r[i];
	}

	int cg_iter = 0;
	rTr = ddot_(&n, r, &inc, r, &inc);
	while (1)
	{
		fun_obj->Hv(d, Hd);
                double dHd = ddot_(&n, d, &inc, Hd, &inc);
                if(dHd <= epsilon*ddot_(&n, d, &inc, d, &inc))
                {
                    if(cg_iter == 0)
                        for (i=0; i<n; i++)
                            s[i] = -g[i];
                    info("CG procedure terminated because the additional termination criterion is satisfied with epsilon = %5.3e \n", epsilon);
                    break;
                }
                cg_iter++;
                alpha = rTr/dHd;
                daxpy_(&n, &alpha, d, &inc, s, &inc);
                alpha = -alpha;
                daxpy_(&n, &alpha, Hd, &inc, r, &inc);

                new_Q = -0.5*(ddot_(&n, s, &inc, r, &inc) - ddot_(&n, s, &inc, g, &inc));
                if (stopping(n, param, g, r, inc, cg_iter, new_Q, Q, Hd, beta) == 1) //the last two variables are not used in the non-preconditioned case, using dummies
                    break;

		rnewTrnew = ddot_(&n, r, &inc, r, &inc);
		beta = rnewTrnew/rTr;
		dscal_(&n, &beta, d, &inc);
		daxpy_(&n, &one, r, &inc, d, &inc);
		rTr = rnewTrnew;
                Q = new_Q;
	}

	delete[] d;
	delete[] Hd;

	return(cg_iter);
}

int TRON::pcg(double *g, double *M, double *s, double *r, const parameter *param)
{
	int i, inc = 1;
	int n = fun_obj->get_nr_variable();
	double one = 1;
	double *d = new double[n];
	double *Hd = new double[n];
	double zTr, znewTrnew, alpha, beta;
	double *z = new double[n];

        double new_Q, Q = 0., epsilon = 0.0001;

	for (i=0; i<n; i++)
	{
		s[i] = 0;
		r[i] = -g[i];
		z[i] = r[i] / M[i];
		d[i] = z[i];
	}

	zTr = ddot_(&n, z, &inc, r, &inc);
	int cg_iter = 0;

        double Minv_dot_g = sqrt(ddot_(&n, z, &inc, r, &inc));

	while (1)
	{
                fun_obj->Hv(d, Hd);
                double dHd = ddot_(&n, d, &inc, Hd, &inc);
                if(dHd <= epsilon*ddot_(&n, d, &inc, d, &inc))
                {
                    if(cg_iter == 0)
                        for (i=0; i<n; i++)
                            s[i] = -g[i];
                    info("CG procedure terminated because the additional termination criterion is satisfied with epsilon = %5.3e \n", epsilon);
                    break;
                }
		cg_iter++;

		alpha = zTr/dHd;
		daxpy_(&n, &alpha, d, &inc, s, &inc);
		alpha = -alpha;
		daxpy_(&n, &alpha, Hd, &inc, r, &inc);

                new_Q = -0.5*(ddot_(&n, s, &inc, r, &inc) - ddot_(&n, s, &inc, g, &inc));
		for (i=0; i<n; i++)
			z[i] = r[i] / M[i];
		znewTrnew = ddot_(&n, z, &inc, r, &inc);

                if (stopping(n, param, g, r, inc, cg_iter, new_Q, Q, z, Minv_dot_g) == 1)
                    break;
		beta = znewTrnew/zTr;
		dscal_(&n, &beta, d, &inc);
		daxpy_(&n, &one, z, &inc, d, &inc);
		zTr = znewTrnew;
                Q = new_Q;

	}

	delete[] d;
	delete[] Hd;
	delete[] z;

	return(cg_iter);
}



double TRON::norm_inf(int n, double *x)
{
	double dmax = fabs(x[0]);
	for (int i=1; i<n; i++)
		if (fabs(x[i]) >= dmax)
			dmax = fabs(x[i]);
	return(dmax);
}

void TRON::set_print_string(void (*print_string) (const char *buf))
{
	tron_print_string = print_string;
}
